/***************************************************************************
 *   Copyright (C) 2003 by jalal (Tim Pushman)                             *
 *   jalal@gnomedia.com                                                    *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 ***************************************************************************/
#include "krefty.h"

#include <kkeydialog.h>
#include <kfiledialog.h>
#include <kconfig.h>
#include <kurl.h>

#include <kedittoolbar.h>

#include <kaction.h>
#include <kstdaction.h>

#include <klibloader.h>
#include <kmessagebox.h>
#include <kstatusbar.h>
#include <kcombobox.h>
#include <klocale.h>

// used for testing...
#include <kstddirs.h>
#include <kglobalsettings.h>

#include <kdebug.h>
#include <khtml_part.h>
//#include <dom/html_document.h>
//#include <dom/dom_string.h>
#include <kio/netaccess.h>          // for file copy

#include <qdom.h>
#include <qdir.h>

#include <qptrlist.h>

#include "krefty_part.h"

const int WIN_STYLE = Qt::WType_TopLevel|Qt::WStyle_Customize|Qt::WStyle_DialogBorder;

//const int WIN_STYLE = Qt::WType_TopLevel|Qt::WStyle_Customize|Qt::WStyle_Title;
//|Qt::WStyle_NoBorder|Qt::WStyle_Title|Qt::WStyle_SysMenu|Qt::WStyle_MinMax;

const int COMBO_ID = 2001;
const QString sDataSubDir      = "data/";
const QString sIntroFile       = "krefty-intro.xml";
const QString sKreftyCSS       = "krefty.css";
const QString sXSLFile         = "krefty.xsl";
const QString sDTDFile         = "krefty.dtd";
const QString sKreftySheetType = "*.xml";
const QString sAppTitle        = "Krefty: ";
const QString sNotePNG         = "note.png";

class KreftySheet
{
public:
    KreftySheet( const QString& n, const KURL& u ) : m_name(n),m_url(u) {}
    QString name() { return m_name; }
    KURL url() { return m_url; }
private:
    KreftySheet(){};
    QString m_name;  // a user friendly name
    KURL    m_url;   // where the sheet is found
};

class KreftySheetList
{
typedef QPtrList<KreftySheet> KreftyList;
public:
    KreftySheetList(){ list.setAutoDelete( true ); }
    ~KreftySheetList(){ list.clear(); }
    void append( const KreftySheet* sheet ) { list.append( sheet ); }
    KURL getURLForSheet( const QString& name );
    QString getNameFromURL( KURL url );
    QStringList getStringList();
    int count() { return list.count(); }
private:
    KreftyList list;
};

Krefty::Krefty()
    : KParts::MainWindow( 0L, "krefty", WIN_STYLE ),m_dataDir("")
{
    kdDebug() << "Krefty::Krefty:: constructor" << endl;
    bool startUpOK = false;

    // check whether the system is setup so that we can use it
    startUpOK = checkSetup();

    // readProperties doesn't seem to get called automatically as
    // it should, so I do it manually.
    readProperties(kapp->config());

    if( setupSheetList() <= 0 ) {
        kdDebug() << "*** ERROR *** Krefty::Krefty(): Could not setup sheet list." << endl;
        KMessageBox::error( this, i18n("The list of krefty files could not be created. Quitting..."));
        startUpOK = false;
        kapp->quit();
        return;
    }

    // set the shell's ui resource file
    setXMLFile("krefty_shell.rc");

    // then, setup our actions
    setupActions();

    // add a combo box to the toolbar and fill it with a sheet listing.
    setUpToolBar();

    // this routine will find and load our Part.  it finds the Part by
    // name which is a bad idea usually.. but it's alright in this
    // case since our Part is made for this Shell
    KLibFactory *factory = KLibLoader::self()->factory("libkreftypart");
    if (factory)
    {
        QString arg( "xslFile=" + m_xslFile );

        // now that the Part is loaded, we cast it to a Part to get
        // our hands on it
        m_part = static_cast<KParts::ReadOnlyPart*>(factory->create(this,
                                "krefty_part", "KParts::ReadOnlyPart", QStringList( arg ) ));

        if (m_part)
        {
            // tell the KParts::MainWindow that this is indeed the main widget
            setCentralWidget(m_part->widget());

            // and integrate the part's GUI with the shell's
            createGUI(m_part);
            setFixedWidth( m_part->widget()->width() /*650*/ );
        }
    } else {
        // if we couldn't find our Part, we exit since the Shell by
        // itself can't do anything useful
        KMessageBox::error(this, i18n("Could not find our Part!"));
        kapp->quit();
        // we return here, cause kapp->quit() only means "exit the
        // next time we enter the event loop...
        return;
    }

    // apply the saved mainwindow settings, if any, and ask the mainwindow
    // to automatically save settings if changed: window size, toolbar
    // position, icon size, etc.
    setAutoSaveSettings();

    kdDebug() << "Krefty::Krefty: using currentURL: " << m_currentURL.url() << endl;
    if( startUpOK && !(m_currentURL.url().isEmpty() || !m_currentURL.isValid()) ) {
        openSheet( m_currentURL );
    } else {
        // error in setup somewhere.
        KMessageBox::error( this, i18n("There was an error starting up Krefty. Quitting..."));
        startUpOK = false;
        kapp->quit();
        return;
    }
    m_part->widget()->setFocus();

    kdDebug() << "Krefty::Krefty:: constructor completed." << endl;
}

Krefty::~Krefty()
{
    delete m_sheetList;
}

void Krefty::openSheet( KURL url )
{
    //kdDebug() << "Krefty::openSheet: with url: " << url.url() << endl;
    m_part->openURL( url );
    QString name = m_sheetList->getNameFromURL( url );
    setComboTo( name );
    setPlainCaption( sAppTitle + name );
}

void Krefty::quit()
{
    // saveProperties doesn't seem to get called automatically :(
    // so i override the quit() and save them here.
    saveProperties(kapp->config());
    kapp->quit();
}

void Krefty::load(const KURL& url)
{
    openSheet( url );
}

// @todo move this into the krefty_part to keep HTML processing together?
int Krefty::setupSheetList()
{
    m_sheetList = new KreftySheetList();
    QDir d( m_dataDir, sKreftySheetType );
    QFileInfoList* entries = const_cast<QFileInfoList*>(d.entryInfoList());
    if( !entries || entries->isEmpty() ) {
        kdDebug() << "*** Krefty::setupSheetList: List of files is empty in " << m_dataDir << endl;
        return 0;
    }
    QFileInfo* f = entries->first();

    // TODO: use SAX parsing to find title element and then stop.
    //       the way it is, is not very efficient as we read in a whole DOM just to find one element.
    QDomDocument doc;
    for( ; f; f = entries->next() ) {
        KURL fname( m_dataDir + f->fileName());
        QFile file( m_dataDir + f->fileName() );
        QString errMsg; int errLine=0, errCol=0;
        if( !file.open( IO_ReadOnly ) ) {
            kdDebug() << "*** Krefty::setupSheetList(): error opening file: " << fname.url() << endl;
            continue;
        }
        if( !doc.setContent( &file, false, &errMsg, &errLine, &errCol ) ) {
            kdDebug() << "*** Krefty::setupSheetList(): bad XML: " << errMsg << " at " << errLine << ":" << errCol << " in " << fname.url() << endl;
            file.close();
            continue;
        }
        file.close();
        if( doc.doctype().name() != "krefty" ){
            kdDebug() << "*** Krefty::setupSheetList(): XML Document is not the correct document type: " << fname.url() << endl;
            continue;
        }
        QDomNodeList nl = doc.elementsByTagName( "title" );
        if(nl.count() <= 0 ) {
            kdDebug() << "*** Krefty::setupSheetList(): Empty node list in file: " << fname.url() << endl;
            continue;
        }
        QDomNode node = nl.item( 0 );
        QString name = node.firstChild().nodeValue();
        if( !name ) {
            kdDebug() << "*** Krefty::setupSheetList(): no title node found in: " << fname.url() << endl;
            continue;
        }
        kdDebug() << "Krefty::setupSheetList: found: " << name << "\n\tat " << fname.url() << endl;
        m_sheetList->append( new KreftySheet( name, fname ) );
    }
    return m_sheetList->count();
}

void Krefty::setupActions()
{
    KStdAction::quit(this, SLOT(quit()), actionCollection());
    new KAction( i18n( "&Refresh List" ), "reload",
                       0,
                       this, SLOT( slotRefreshList() ),
                       actionCollection(), "refresh_list" );

    m_toolbarAction = KStdAction::showToolbar(this, SLOT(optionsShowToolbar()), actionCollection());
}

void Krefty::setUpToolBar()
{
    //kdDebug() << "Krefty::setUpToolBar():" << endl;
    QStringList list = m_sheetList->getStringList();
    if( list.size() <= 0 ) {
        kdDebug() << "*** ERROR *** Krefty::setUpToolBar(): no entries in string list!" << endl;
    } else {
        QString tip( i18n("Select a krefty sheet from here") );
        toolBar()->clear();
        toolBar()->insertCombo(list,                                    // QStringList
                               COMBO_ID,                                // ID
                               false,                                   // writable?
                               SIGNAL(activated(const QString&) ),      // signal
                               this,                                    // receiver
                               SLOT(slotComboChanged(const QString&)),  // slot
                               true,                                    // enabled?
                               tip,                                     // tooltip text
                               HTML_VIEW_WIDTH - 20,                    // width
                               -1 );                                    // index
    }
}

void Krefty::slotRefreshList()
{
    QString name = toolBar()->getCombo( COMBO_ID )->currentText();

    delete m_sheetList;
    setupSheetList();
    toolBar()->clearCombo( COMBO_ID );
    toolBar()->insertComboList( COMBO_ID, m_sheetList->getStringList(), 0 );
    setComboTo( name );
}

void Krefty::setComboTo( QString sheetName )
{
    KComboBox* combo = toolBar()->getCombo( COMBO_ID );
    combo->setCurrentItem( sheetName );
}

void Krefty::slotComboChanged( const QString& name )
{
    // index is 0 based
    KURL url = m_sheetList->getURLForSheet( name );
    if( url.isEmpty() ) {
        kdDebug() << "**** Krefty::slotComboChanged(): empty URL returned from getURLForSheet() with " << name << endl;
    } else {
        //kdDebug() << "Krefty::slotComboChanged(): with " << name << " at " << url.url() << endl;
        m_currentURL = url;
        openSheet( url );
    }
    m_part->widget()->setFocus();
}

void Krefty::saveProperties(KConfig* config )
{
    // the 'config' object points to the session managed
    // config file.  anything you write here will be available
    // later when this app is restored
    //kdDebug() << "+++ Krefty::saveProperties:\n\t" << m_currentURL.url() << "\n\t" << m_dataDir << "\n\t" << m_xslFile << endl;
    config->setGroup( "Sheets" );
    config->writeEntry( "lastSheet", m_currentURL.url() );
    config->writeEntry( "dataDir", m_dataDir );
    config->writeEntry( "xslFile", m_xslFile );

    // this is a bit of a hack to override the minimum font size so that we can specify
    // it in the CSS style sheet.
    config->setGroup( "HTML Settings" );
    config->writeEntry( "MinimumFontSize", 6 );
}

void Krefty::readProperties(KConfig* config)
{
    // the 'config' object points to the session managed
    // config file.  this function is automatically called whenever
    // the app is being restored.  read in here whatever you wrote
    // in 'saveProperties'
    config->setGroup( "Sheets" );
    QString lastSheet = config->readEntry( "lastSheet", m_currentURL.path() );
    m_dataDir = config->readEntry( "dataDir", m_dataDir );
    m_xslFile = config->readEntry( "xslFile", sXSLFile );
    if( !lastSheet.isEmpty() ) { // take care of the situation where startup may have failed without writing a lastSheet
        m_currentURL = KURL(lastSheet);
    }
    //kdDebug() << "+++ Krefty::readProperties:\n\t" << lastSheet << "\n\t" << m_dataDir << "\n\t" << m_xslFile << endl;
}

void Krefty::optionsShowToolbar()
{
    // this is all very cut and paste code for showing/hiding the
    // toolbar
    if (m_toolbarAction->isChecked())
        toolBar()->show();
    else
        toolBar()->hide();
}

void Krefty::fileOpen()
{
    // this slot is called whenever the File->Open menu is selected,
    // the Open shortcut is pressed (usually CTRL+O) or the Open toolbar
    // button is clicked
    KURL url =
        KFileDialog::getOpenURL( QString::null, QString::null, this );

    if (url.isEmpty() == false)
    {
        // About this function, the style guide (
        // http://developer.kde.org/documentation/standards/kde/style/basics/index.html )
        // says that it should open a new window if the document is _not_
        // in its initial state.  This is what we do here..
        if ( m_part->url().isEmpty() /*&& ! m_part->isModified()*/ )
        {
            // we open the file in this window...
            load( url );
        }
        else
        {
            // we open the file in a new window...
            Krefty* newWin = new Krefty;
            newWin->load( url );
            newWin->show();
        }
    }
}

bool Krefty::checkSetup()
{
    bool retval = false;
    //QString ifile = sDataSubDir + sIntroFile;
    QString locData = locateLocal( "appdata", sDataSubDir + sIntroFile );
    kdDebug() << "+++ Krefty::checkSetup(): locData: " << locData << endl;
    KURL userData( KURL(locData).directory(false) );
    m_dataDir = userData.path(false); // false = keep trailing slash
    m_currentURL = locData;           // useful default for now.
    kdDebug() << "Krefty::checkSetup: dataDir is " << m_dataDir << endl;
    kdDebug() << "Krefty::checkSetup: currentURL is " << m_currentURL.url() << endl;

    if( QFile::exists( locData ) ) {
        retval = true;
    } else {
        KConfig* config = kapp->config();
        config->setGroup( "HTML Settings" );
        config->writeEntry( "MinimumFontSize", 6 );
        return copyStartUpFiles( userData );
    }
    return retval;
}

//
// TODO: this needs to copy the xsl, dtd and xml files
//
bool Krefty::copyStartUpFiles( KURL& userData )
{
    // create the directory, populate it with files taken from the
    // directory $KDEDIR/share/apps/krefty/data
    //
    bool retval = false;
    QString sysData = KGlobal::instance()->dirs()->findResourceDir("appdata", sDataSubDir + sIntroFile );
    if( !sysData ){
        kdDebug() << "*** Krefty::copyStartUpFiles(): Cannot find the original files." << endl;
    }
    kdDebug() << "+++ Krefty::copyStartUpFiles(): user data at: " << userData.path() << endl;
    kdDebug() << "+++ Krefty::copyStartUpFiles(): system data at: " << sysData << endl;
    if( sysData ) {
        if( QFile::exists(sysData) ) {
            userData.setFileName( sIntroFile );
            if( !KIO::NetAccess::copy( KURL(sysData + sDataSubDir + sIntroFile), userData ) ) {
                kdDebug() << "Krefty::copyStartUpFiles(): file copy failed! " << (sysData + sDataSubDir + sIntroFile) << " to " << userData.url() << endl;
            } else {
                m_currentURL = userData;
                retval = true;
            }
            userData.setFileName( sKreftyCSS );
            if( retval && !KIO::NetAccess::copy( KURL(sysData + sDataSubDir + sKreftyCSS), userData ) ) {
                kdDebug() << "*** Krefty::copyStartUpFiles(): file copy failed! " << (sysData + sDataSubDir + sKreftyCSS) << " to " << userData.url() << endl;
                retval = false;
            }
            userData.setFileName( sXSLFile );
            if( retval && !KIO::NetAccess::copy( KURL(sysData + sDataSubDir + sXSLFile), userData ) ) {
                kdDebug() << "*** Krefty::copyStartUpFiles(): file copy failed! " << (sysData + sDataSubDir + sXSLFile) << " to " << userData.url() << endl;
                retval = false;
            }
            userData.setFileName( sDTDFile );
            if( retval && !KIO::NetAccess::copy( KURL(sysData + sDataSubDir + sDTDFile), userData ) ) {
                kdDebug() << "*** Krefty::copyStartUpFiles(): file copy failed! " << (sysData + sDataSubDir + sDTDFile) << " to " << userData.url() << endl;
                retval = false;
            }
            userData.setFileName( sNotePNG );
            if( retval && !KIO::NetAccess::copy( KURL(sysData + sDataSubDir + sNotePNG), userData ) ) {
                kdDebug() << "*** Krefty::copyStartUpFiles(): file copy failed! " << (sysData + sDataSubDir + sNotePNG) << " to " << userData.url() << endl;
                retval = false;
            }
        }
    } else {
        kdDebug() << "*** ERROR: Krefty::copyStartUpFiles(): system data not found: " << sysData << endl;
    }
    return retval;
}

/*****************************************************************************************************
 * KreftySheetList and KreftySheet methods                                                           *
 *****************************************************************************************************/
QStringList KreftySheetList::getStringList()
{
    //kdDebug() << "KreftySheetList::getStringList(): started..." << endl;
    KreftySheet* s;
    QStringList l;
    for( s = list.first(); s; s = list.next() ) {
        kdDebug() << "KreftySheetList::getStringList():name: " << s->name() << endl;
        l.append( s->name() );
    }
    return l;
}

KURL KreftySheetList::getURLForSheet( const QString& name )
{
    KURL url;
    if( !list.isEmpty() ) {
        KreftySheet* k = list.first();
        for( ; k; k = list.next() ) {
            if( k->name() == name ) {
                url = k->url();
                break;
            }
        }
    }
    return url;
}

QString KreftySheetList::getNameFromURL( KURL url )
{
    QString name;
    if( !list.isEmpty() ) {
        KreftySheet* k = list.first();
        for( ; k; k = list.next() ) {
            if( k->url() == url ) {
                name = k->name();
                break;
            }
        }
    }
    return name;
}


#include "krefty.moc"

